#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <ctype.h>

typedef struct NODESTRUCT
{
	unsigned int id ;
	unsigned int depth ;
	unsigned int node_type ;
	unsigned int data_size ;
	unsigned char *data ;
	unsigned short numchildren ;
	struct NODESTRUCT *children ;
	struct NODESTRUCT *parent ;
} node ;

node root ;
FILE *dtbfile ;
unsigned char dtb_header[7] ;
unsigned int filepos ;
unsigned int childnum_depth[50] ;
unsigned int dtb_loaded = 0 ;

void* safe_malloc( size_t memsize )
{
	void *ptr ;

	ptr = malloc( memsize ) ;

	if ( ptr == NULL )
	{
		printf( "Could not allocate %d bytes of data - aborting\n", memsize ) ;
		exit( 1 ) ;
	}

	return ptr ;
}

void read_branch( node *parent, int depth )
{
	int i ;

	for ( i = 0 ; i < parent->numchildren ; i++ )
	{
		node *child = &((parent->children)[i]) ;

		child->parent = parent ;
		child->depth = depth ;

		fread( &(child->node_type), 1, 4, dtbfile ) ;
		filepos += 4 ;

		switch ( child->node_type )
		{
			case 0x0 :
			case 0x1 :
			case 0x6 :
			case 0x8 :
			case 0x9 :
			{
				//4 bytes of data
				child->data_size = 4 ;
				child->data = safe_malloc( 4 ) ;
				fread( child->data, 1, 4, dtbfile ) ;
				filepos += 4 ;

				//printf( "read static 4\n" ) ;
				break ;
			}
			case 0x2 :
			case 0x5 :
			case 0x7 :
			case 0x12:
			case 0x20:
			case 0x21:
			case 0x22:
			case 0x23:
			{
				//variable size data - first 4 bytes is size of data, next n bytes is data
				fread( &(child->data_size), 1, 4, dtbfile ) ;
				child->data = safe_malloc( child->data_size + 1) ;
				child->data[child->data_size] = 0 ;
				fread( child->data, 1, child->data_size, dtbfile ) ;
				filepos += 4+child->data_size ;
				//printf( "read var %d\n", child->data_size ) ;
				break ;
			}
			case 0x10:
			case 0x11:
			case 0x13:
			{
				//branch
				fread( &(child->numchildren), 1, 2, dtbfile ) ;
				child->children = (node*)safe_malloc( sizeof(node) * child->numchildren ) ;
				memset( child->children, 0, sizeof(node) * child->numchildren ) ;
				fread( &(child->id), 1, 4, dtbfile ) ;
				filepos += 6 ;

				//printf( "read branch id\n" ) ;
				read_branch( child, depth+1  ) ;
				break ;
			}
			default :
			{
				//unknown
				printf( "unknown data type %d filepos %d\n", child->node_type, filepos ) ;
				break ;
			}
		}
	}
}

void write_branch( node *parent )
{
	int i ;

	for ( i = 0 ; i < parent->numchildren ; i++ )
	{
		node *child = &((parent->children)[i]) ;

		fwrite( &(child->node_type), 1, 4, dtbfile ) ;
		filepos += 4 ;

		switch ( child->node_type )
		{
			case 0x0 :
			case 0x1 :
			case 0x6 :
			case 0x8 :
			case 0x9 :
			{
				//4 bytes of data
				fwrite( child->data, 1, 4, dtbfile ) ;
				filepos += 4 ;

				//printf( "read static 4\n" ) ;
				break ;
			}
			case 0x2 :
			case 0x5 :
			case 0x7 :
			case 0x12:
			case 0x20:
			case 0x21:
			case 0x22:
			case 0x23:
			{
				//variable size data - first 4 bytes is size of data, next n bytes is data
				fwrite( &(child->data_size), 1, 4, dtbfile ) ;
				fwrite( child->data, 1, child->data_size, dtbfile ) ;
				filepos += 4+child->data_size ;
				//printf( "read var %d\n", child->data_size ) ;
				break ;
			}
			case 0x10:
			case 0x11:
			case 0x13:
			{
				//branch
				fwrite( &(child->numchildren), 1, 2, dtbfile ) ;
				fwrite( &(child->id), 1, 4, dtbfile ) ;
				filepos += 6 ;

				//printf( "read branch id\n" ) ;
				write_branch( child  ) ;
				break ;
			}
			default :
			{
				//unknown
				printf( "unknown data type %d filepos %d\n", child->node_type, filepos ) ;
				break ;
			}
		}
	}
}

void printid( int depth)
{
	int i  ;

	for ( i = 0 ; i<=depth ; i++ )
	{
		if ( i == depth )
			printf( "%d", childnum_depth[i] ) ;
		else
			printf( "%d.", childnum_depth[i] ) ;
	}
}

void print_branch( node *parent, int depth )
{
	int i ;
	char *spacer ;

	spacer = safe_malloc( (depth*3) + 1 ) ;
	memset( spacer, ' ', depth*3 ) ;
	spacer[depth*3] = 0 ;

	printf( "%sBranch ID : %d (%d children)\n", spacer, parent->id, parent->numchildren ) ;

	for ( i = 0 ; i < parent->numchildren ; i++ )
	{
		node *child = &((parent->children)[i]) ;
		childnum_depth[depth] = i ;

		switch ( child->node_type )
		{
			case 0x0 :
			case 0x6 :
			case 0x8 :
			case 0x9 :
			{
				//0 = int, 6,8,9 = unknown, print as int
				int data = *( (int*)child->data ) ;
				printf( "%s   Int (%04d)         : %d (ID ", spacer, child->node_type, data ) ;
				printid( depth ) ;
				printf( " )\n" ) ;
				break ;
			}
			case 0x1 :
			{
				//1 = float
				float data = *( (float*)child->data ) ;
				printf( "%s   Float (%04d)       : %f (ID ", spacer, child->node_type, data ) ;
				printid( depth ) ;
				printf( " )\n" ) ;

				break ;
			}
			case 0x2 :
			case 0x5 :
			case 0x7 :
			case 0x12:
			case 0x20:
			case 0x21:
			case 0x22:
			case 0x23:
			{
				//variable size data - first 4 bytes is size of data, next n bytes is data
				printf( "%s   String (%04d,%04d) : %s (ID ", spacer, child->node_type, child->data_size, child->data ) ;
				printid( depth ) ;
				printf( " )\n" ) ;
				break ;
			}
			case 0x10:
			case 0x11:
			case 0x13:
			{
				//branch
				print_branch( child, depth+1  ) ;
				break ;
			}
		}
	}
}

int getnextnum( char *str, int *pos, int *num )
{
	*num = 0 ;

	if ( *pos >= strlen( str ) )
		return 0 ;

	while ( ( str[*pos] != '.' ) && ( str[*pos] != 0 ) )
	{
		if ( isdigit( str[*pos] ) )
		{
			*num = *num * 10 + (str[*pos]-48) ;
		}
        else
            return 0 ;

		(*pos)++ ;
	}

	if ( str[*pos] == '.' )
		(*pos)++ ;

	return 1 ;
}

node* dtb_lookup_node( char *str )
{
	int pos = 0 ;
	int num = 0 ;
	node *parent = &root ;

	while( getnextnum( str, &pos, &num ) )
	{
//		printf( "found num %d\n", num ) ;
		if ( num < parent->numchildren )
		{
			parent = &(parent->children[num]) ;
		}
        else
            return &root ;
	}

    if ( str[pos] == 0 )
        return parent ;
    else
        return &root ;

}

int dtb_get_string( char *str, char **val )
{
    node *entry ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        *val = (char*)(entry->data) ;
        return 0 ;
    }
}

int dtb_get_int( char *str, int *val )
{
    node *entry ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        *val = *(int*)(entry->data) ;
        return 0 ;
    }
}

int dtb_get_float( char *str, float *val )
{
    node *entry ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        *val = *(float*)(entry->data) ;
        return 0 ;
    }
}

int dtb_set_string( char *str, char *val )
{
    node *entry ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        free( entry->data ) ;
        entry->data = safe_malloc( strlen(val) + 1 );
        memcpy( entry->data, val, strlen(val) ) ;
        entry->data[strlen(val)] = 0 ;
        entry->data_size = strlen(val) ;
        return 0 ;
    }
}

int dtb_add_string_after( char *str, char *val )
{
    node *entry ;
    node *newentry ;
    node *newchildren ;
    node *parent ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        newchildren = (node*)safe_malloc( (entry->parent->numchildren + 1) * sizeof(node) ) ;

        memcpy( newchildren, entry->parent->children, sizeof(node)*(entry->parent->numchildren) ) ;

        newentry = newchildren + ( entry->parent->numchildren ) ;

        newentry->id = 0 ;
        newentry->depth = entry->depth ;
        newentry->node_type = 5 ;
        newentry->data_size = strlen(val) ;
        newentry->data = (unsigned char*)safe_malloc( strlen(val) + 1 ) ;
        memcpy( newentry->data, val, strlen(val) ) ;
        newentry->data[newentry->data_size] = 0 ;
        newentry->numchildren = 0 ;
        newentry->children = NULL ;
        newentry->parent = entry->parent ;

        parent = entry->parent ;

        free( entry->parent->children ) ;
        parent->children = newchildren ;
        parent->numchildren = parent->numchildren + 1 ;

        return 0 ;
    }
}

int dtb_add_float_after( char *str, float val )
{
    node *entry ;
    node *newentry ;
    node *newchildren ;
    node *parent ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        newchildren = (node*)safe_malloc( (entry->parent->numchildren + 1) * sizeof(node) ) ;

        memcpy( newchildren, entry->parent->children, sizeof(node)*(entry->parent->numchildren) ) ;

        newentry = newchildren + ( entry->parent->numchildren ) ;

        newentry->id = 0 ;
        newentry->depth = entry->depth ;
        newentry->node_type = 1 ;
        newentry->data_size = 4 ;
        newentry->data = (unsigned char*)safe_malloc( 4 ) ;
        *( (float*)newentry->data ) = val ;
        newentry->numchildren = 0 ;
        newentry->children = NULL ;
        newentry->parent = entry->parent ;

        parent = entry->parent ;

        free( entry->parent->children ) ;
        parent->children = newchildren ;
        parent->numchildren = parent->numchildren + 1 ;

        return 0 ;
    }
}
int dtb_set_int( char *str, int val )
{
    node *entry ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        memcpy( entry->data, &val, sizeof(int)) ;
        return 0 ;
    }
}

int dtb_set_float( char *str, float val )
{
    node *entry ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        memcpy( entry->data, &val, sizeof(float)) ;
        return 0 ;
    }
}

int dtb_add_int_after( char *str, int val )
{
    node *entry ;
    node *newentry ;
    node *newchildren ;
    node *parent ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        newchildren = (node*)safe_malloc( (entry->parent->numchildren + 1) * sizeof(node) ) ;

        memcpy( newchildren, entry->parent->children, sizeof(node)*(entry->parent->numchildren) ) ;

        newentry = newchildren + ( entry->parent->numchildren ) ;

        newentry->id = 0 ;
        newentry->depth = entry->depth ;
        newentry->node_type = 0 ;
        newentry->data_size = 4 ;
        newentry->data = (unsigned char*)safe_malloc( 4 ) ;
        *( (int*)newentry->data ) = val ;
        newentry->numchildren = 0 ;
        newentry->children = NULL ;
        newentry->parent = entry->parent ;

        parent = entry->parent ;

        free( entry->parent->children ) ;
        parent->children = newchildren ;
        parent->numchildren = parent->numchildren + 1 ;

        return 0 ;
    }
}


int dtb_fixup_track( char *str )
{
    node *entry ;
    node *newentry ;
    node *newchildren ;
    node *parent ;
    node *child ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        entry = entry->parent->children+1 ;

        if ( entry->numchildren == 1 )
        {
            newchildren = (node*)safe_malloc( (entry->numchildren + 1) * sizeof(node) ) ;
            memset( newchildren, 0, sizeof(node) * (entry->numchildren + 1) ) ;
            memcpy( newchildren, entry->children, sizeof(node)*(entry->numchildren) ) ;
            newentry = newchildren + ( entry->numchildren ) ;

            newentry->id = entry->id ;
            newentry->depth = entry->depth+1 ;
            newentry->node_type = entry->node_type ;
            newentry->data_size = 0 ;
            newentry->data = NULL ;
            newentry->numchildren = 2 ;
            newentry->children = (node*)safe_malloc( (newentry->numchildren) * sizeof(node) ) ;
            memset( newentry->children, 0, sizeof(node) * (newentry->numchildren) ) ;
            newentry->parent = entry ;

            free( entry->children ) ;
            entry->children = newchildren ;
            entry->numchildren = entry->numchildren + 1 ;

            newchildren = newentry->children ;

            child = newchildren+0 ;

            child->id = 0 ;
            child->depth = entry->depth+1 ;
            child->node_type = 5 ;
            child->data_size = 4 ;
            child->data = (unsigned char*)safe_malloc( 5 ) ;
            memcpy( child->data, "bass", 4 ) ;
            child->data[child->data_size] = 0 ;
            child->numchildren = 0 ;
            child->children = NULL ;
            child->parent = entry ;


            newentry = newchildren+1 ;

            newentry->id = entry->id ;
            newentry->depth = entry->depth+1 ;
            newentry->node_type = entry->node_type ;
            newentry->data_size = 0 ;
            newentry->data = NULL ;
            newentry->numchildren = 2 ;
            newentry->children = (node*)safe_malloc( (newentry->numchildren) * sizeof(node) ) ;
            memset( newentry->children, 0, sizeof(node) * (newentry->numchildren) ) ;
            newentry->parent = entry ;

            child = newentry->children+0 ;

            child->id = 0 ;
            child->depth = newentry->depth+1 ;
            child->node_type = 0 ;
            child->data_size = 4 ;
            child->data = (unsigned char*)safe_malloc( 4 ) ;
            *((int*)(child->data)) = 4 ;
            child->numchildren = 0 ;
            child->children = NULL ;
            child->parent = newentry ;

            child = newentry->children+1 ;

            child->id = 0 ;
            child->depth = newentry->depth+1 ;
            child->node_type = 0 ;
            child->data_size = 4 ;
            child->data = (unsigned char*)safe_malloc( 4 ) ;
            *((int*)(child->data)) = 5 ;
            child->numchildren = 0 ;
            child->children = NULL ;
            child->parent = newentry ;


        }
        else
        {
            parent = entry->children+1 ;
            entry = ((parent->children)+1) ;

            if ( entry->node_type == 0 )
            {
                entry->id = parent->id ;
                entry->depth = parent->depth+1 ;
                entry->node_type = parent->node_type ;
                entry->data_size = 0 ;
                entry->data = NULL ;
                entry->numchildren = 2 ;
                entry->children = (node*)safe_malloc( (entry->numchildren) * sizeof(node) ) ;
                memset( entry->children, 0, sizeof(node) * (entry->numchildren) ) ;
                entry->parent = parent ;


                child = entry->children+0 ;

                child->id = 0 ;
                child->depth = entry->depth+1 ;
                child->node_type = 0 ;
                child->data_size = 4 ;
                child->data = (unsigned char*)safe_malloc( 4 ) ;
                *((int*)(child->data)) = 4 ;
                child->numchildren = 0 ;
                child->children = NULL ;
                child->parent = entry ;

                child = entry->children+1 ;

                child->id = 0 ;
                child->depth = entry->depth+1 ;
                child->node_type = 0 ;
                child->data_size = 4 ;
                child->data = (unsigned char*)safe_malloc( 4 ) ;
                *((int*)(child->data)) = 5 ;
                child->numchildren = 0 ;
                child->children = NULL ;
                child->parent = entry ;



            }
        }
    }


    return 0 ;
}

int dtb_fixup_vols( char *str )
{
    char nodename[20] ;
    char lastnodename[20] ;
    node *entry ;
    int i ;
    int len ;

    strcpy( nodename, str ) ;

    len = strlen(nodename) ;
    nodename[len+2]   = 0 ;
    nodename[len-1] = '1' ;
    nodename[len]   = '.' ;

    for ( i = 0 ; i < 6 ; i++ )
    {
        nodename[len+1]   = '0' + i ;
        entry = dtb_lookup_node( nodename ) ;

        if ( entry == &root )
        {
            dtb_add_float_after( lastnodename, -2.5f ) ;
        }
        else
        {
            dtb_set_float( nodename, -2.5f ) ;
        }

        strcpy( lastnodename, nodename ) ;

    }



    return 0 ;
}

int dtb_fixup_pans( char *str )
{
    char nodename[20] ;
    char lastnodename[20] ;
    node *entry ;
    int i ;
    int len ;

    strcpy( nodename, str ) ;

    len = strlen(nodename) ;

    nodename[len+2]   = 0 ;
    nodename[len-1] = '1' ;
    nodename[len]   = '.' ;

    for ( i = 0 ; i < 6 ; i++ )
    {
        nodename[len+1]   = '0' + i ;
        entry = dtb_lookup_node( nodename ) ;

        if ( entry == &root )
        {
            dtb_add_float_after( lastnodename, i%2 ? 1.0f : -1.0f ) ;
        }
        else
        {
            dtb_set_float( nodename, i%2 ? 1.0f : -1.0f ) ;
        }

        strcpy( lastnodename, nodename ) ;

    }



    return 0 ;
}

int dtb_fixup_cores( char *str )
{
    char nodename[20] ;
    char lastnodename[20] ;
    node *entry ;
    int i ;
    int len ;

    strcpy( nodename, str ) ;

    len = strlen(nodename) ;

    nodename[len+2]   = 0 ;
    nodename[len-1] = '1' ;
    nodename[len]   = '.' ;

    for ( i = 0 ; i < 6 ; i++ )
    {
        nodename[len+1]   = '0' + i ;
        entry = dtb_lookup_node( nodename ) ;

        if ( entry == &root )
        {
            dtb_add_int_after( lastnodename, ( ( i == 2 ) || ( i == 3 ) ) ? 1: -1 ) ;
        }
        else
        {
            dtb_set_int( nodename, ( ( i == 2 ) || ( i == 3 ) ) ? 1: -1 ) ;
        }

        strcpy( lastnodename, nodename ) ;

    }



    return 0 ;
}

int dtb_delete_entry( char *str )
{
    node *entry ;
    node *parent ;
    node *newchildren ;
    node *curr ;
    node *dest ;
    int n ;

    entry = dtb_lookup_node( str ) ;

    if ( entry == &root )
    {
        return 1 ;
    }
    else
    {
        free( entry->data ) ;

        parent = entry->parent ;


        newchildren = (node*)safe_malloc( sizeof(node) * ( parent->numchildren - 1 )) ;

        dest = newchildren ;
        curr = parent->children ;

        for ( n = 0 ; n < parent->numchildren ; n++ )
        {

            if ( curr != entry )
            {
                memcpy( dest, curr, sizeof(node) ) ;
                dest++ ;
            }
            curr++ ;
        }

        free( parent->children ) ;
        parent->children = newchildren ;
        parent->numchildren = parent->numchildren-1 ;

        return 0 ;
    }
}


void dtb_clear_branch( node *parent )
{
	int i ;

	for ( i = 0 ; i < parent->numchildren ; i++ )
	{
		node *child = &((parent->children)[i]) ;


		switch ( child->node_type )
		{
			case 0x0 :
			case 0x1 :
			case 0x6 :
			case 0x8 :
			case 0x9 :
			case 0x2 :
			case 0x5 :
			case 0x7 :
			case 0x12:
			case 0x20:
			case 0x21:
			case 0x22:
			case 0x23:
			{
				free( child-> data ) ;
				break ;
			}
			case 0x10:
			case 0x11:
			case 0x13:
			{
				//branch
				dtb_clear_branch( child ) ;
                free(child->children) ;

				break ;
			}
			default :
			{
				//unknown
				printf( "unknown data type %d filepos %d\n", child->node_type, filepos ) ;
				break ;
			}
		}
	}
}

int dtb_read( char *filename )
{
    if ( dtb_loaded )
    {
        dtb_clear_branch( &root ) ;
        free( root.children ) ;
        root.children = NULL ;
    }

    dtb_loaded = 0 ;

	dtbfile = fopen( filename, "rb" ) ;
	if ( dtbfile==NULL )
	{
		return 1 ;
	}

	fread( dtb_header, 1, 7, dtbfile ) ;

	filepos = 7 ;

	memset( childnum_depth, 0, sizeof(int)*50 ) ;

	root.id = 0 ;
	root.node_type = 0xff ;
	root.data_size = 0 ;
	root.data = NULL ;
	root.numchildren = *((unsigned short*)(dtb_header+1)) ;
	root.parent = NULL ;
	root.children = (node*)safe_malloc( sizeof(node) * root.numchildren ) ;
	root.depth = 0 ;
	memset( root.children, 0, sizeof(node) * root.numchildren ) ;

	read_branch( &root, 1  ) ;

	fclose( dtbfile ) ;

    dtb_loaded = 1 ;

    return 0 ;
}

int dtb_write_file( FILE *file)
{
	dtbfile = file ;
	if ( dtbfile==NULL )
	{
		return 1 ;
	}

	fwrite( dtb_header, 1, 7, dtbfile ) ;

	write_branch( &root ) ;

    return 0 ;
}

int dtb_write( char *filename)
{
	dtbfile = fopen( filename, "wb" ) ;
	if ( dtbfile==NULL )
	{
		return 1 ;
	}

	fwrite( dtb_header, 1, 7, dtbfile ) ;

	write_branch( &root ) ;

	fclose( dtbfile ) ;

    return 0 ;
}


#ifndef GH2MAKER
int main( int argc, char *argv[] )
{

	if ( argc < 2 )
	{
		printf( "usage:  %s <dtbfilename>\n", argv[0] ) ;
		return 1 ;
	}

	dtbfile = fopen( argv[1], "rb" ) ;
	if ( dtbfile==NULL )
	{
		printf( "could not open %s for reading\n", argv[1] ) ;
		return 1 ;
	}

	fread( dtb_header, 1, 7, dtbfile ) ;

	filepos = 7 ;

	memset( childnum_depth, 0, sizeof(int)*50 ) ;

	root.id = 0 ;
	root.node_type = 0xff ;
	root.data_size = 0 ;
	root.data = NULL ;
	root.numchildren = *((unsigned short*)(dtb_header+1)) ;
	root.parent = NULL ;
	root.children = (node*)safe_malloc( sizeof(node) * root.numchildren ) ;
	root.depth = 0 ;
	memset( root.children, 0, sizeof(node) * root.numchildren ) ;

	read_branch( &root, 1  ) ;

	fclose( dtbfile ) ;

	print_branch( &root, 0 ) ;

	dtbfile = fopen( "newfile.dtb", "wb" ) ;
	if ( dtbfile==NULL )
	{
		printf( "could not open newfile.dtb for writing\n" ) ;
		return 1 ;
	}

	fwrite( dtb_header, 1, 7, dtbfile ) ;

	write_branch( &root ) ;

	fclose( dtbfile ) ;

	{
//		node *lnode = lookup_node( "1.8.6.1" ) ;
//		printf( "node data = %s\n",  lnode->data ) ;
	}
	return 0 ;
}
#endif
